/* simulate sample paths geometric brownian motion */
/*
	Copyright 2001 Gunter Winkler
*/

#include "gbm.h"

#define odd(x)  (x & 0x01)

static inline int sign(const double x)
{
	if (x>0.0) return 1;
	if (x<0.0) return -1;
	return 0;
}

double wiener_final(PMTDATA handle, const double mu, const double sigma,
		const double start, const double dt, const int steps)
/* 
	simulate 'steps' steps of width 'dt' of a Wiener process, starting 
	at S(0)='start', with drift 'mu', variance 'sigma'
	result is S(steps*dt)
*/
{
	double S = start;
	double x1,x2;
	double sdt = sqrt(dt);
	int i;
	
	gennormal(handle,x1,x2);
	S += mu * dt + sigma * x1 * sdt;
	if (!odd(steps)) S += mu * dt + sigma * x2 * sdt;
	
	for (i=(steps/2)-1; i>0; --i)
	{
		gennormal(handle,x1,x2);
		S += mu * dt + sigma * x1 * sdt;
		S += mu * dt + sigma * x2 * sdt;
	}
	return S;
}

double gbm_final(PMTDATA handle, const double mu, const double sigma,
		const double start, const double dt, const int steps)
/* 
	simulate 'steps' steps of width 'dt' of a gbm, starting with S(0)='start',
	drift 'mu', variance 'sigma'
	result is S(steps*dt)
*/
{
	double S = wiener_final(handle,mu-sigma*sigma/2.0,sigma,0.0,dt,steps);
	return start*exp(S);
}

#define STEP(x) S += S * (mu * dt + sigma * x * sdt);

double gbm2_final(PMTDATA handle, const double mu, const double sigma,
		const double start, const double dt, const int steps)
/* 
	simulate 'steps' steps of width 'dt' of a Wiener process, starting 
	at S(0)='start', with drift 'mu', variance 'sigma'
	result is S(steps*dt)
*/
{
	double S = start;
	double x1,x2;
	double sdt = sqrt(dt);
	int i;
	
	gennormal(handle,x1,x2);
	STEP(x1)
	if (!odd(steps)) STEP(x2)
	
	for (i=(steps/2)-1; i>0; --i)
	{
		gennormal(handle,x1,x2);
		STEP(x1)
		STEP(x2)
	}
	return S;
}

int gbm_barrier(PMTDATA handle, const double mu, const double sigma,
		double & S, const double dt, const int steps, 
		const double barrier)
/* 
	simulate 'steps' steps of width 'dt' of a Wiener process, starting 
	at given S(0), with drift 'mu', variance 'sigma'
	result is S(steps*dt)
*/
{
	double x1,x2;
	double sdt = sqrt(dt);
	int i,updown;
	
	updown = sign(S-barrier);
	if (updown==0) return 0;
	/* abort if the option starts at the barrier */
	
	gennormal(handle,x1,x2);
	STEP(x1)
	if (sign(S-barrier) != updown) return 1;
	if (!odd(steps)) 
	{			
		STEP(x2)
		if (sign(S-barrier) != updown) return 2;
	}
	
	for (i=(steps/2)-1; i>0; --i)
	{
		gennormal(handle,x1,x2);
		STEP(x1)
		if (sign(S-barrier) != updown) return (steps-(2*i+1));
		STEP(x2)
		if (sign(S-barrier) != updown) return (steps-(2*i));
	}
	return -1;
}

#undef STEP
